type lowimage = {
    l_pixels : string; (* flapped *)
    l_width : int;
    l_height : int;
    l_colors : int;
    l_red : int array;
    l_green : int array;
    l_blue : int array;
    l_free : bool array;
    l_transparent : int
  } 

type image = {
    pixels : string;
    width : int;
    height : int;
    palette : (int * int * int) array 
  } 

exception GifError of string

external install_giferror_exn : exn -> unit
  = "install_giferror_exn"

external loadlow : string -> lowimage
  = "loadGifFile"

(* init of exception *)
let _ = install_giferror_exn (GifError "")

let load_palette file =
  let fileHd = open_in_bin file in
  let pal = Array.create 256 (-1,-1,-1) in
  for i=0 to 255 do
    let red = int_of_string (input_line fileHd) in
    let green = int_of_string (input_line fileHd) in
    let blue = int_of_string (input_line fileHd) in
    pal.(i) <- (red,green,blue)
  done ;
  close_in fileHd;
  pal

let get_nearest pal (r,g,b) =
  let dif = ref (-1) in
  let c = ref (-1) in
  for i = 0 to Array.length pal - 1 do
    let pr, pg, pb = pal.(i) in
    let d = (pr - r)*(pr - r) + (pg - g)*(pg - g) + (pb - b)*(pb - b) in
    if !dif > d or !dif < 0 then begin
      dif := d;
      c := i
    end
  done;
  !c


(* ************************************** *)
(* LOADS A GIF FILES WITH ITS OWN PALETTE *)
(* ************************************** *)
let load name =
  (* use lower interf. *)
  let limg = loadlow name in
  (* translate *)
  let pixels = String.create (limg.l_width * limg.l_height) in
  let trans = Array.create 256 '\000' in
  let palette = Array.create 256 (-1,-1,-1) in
  let j = ref 0 in
  for i = 0 to 256 - 1 do
    if limg.l_free.(i) then () 
    else begin
      trans.(i) <- Char.chr !j;
      palette.(!j) <- limg.l_red.(i), limg.l_green.(i), limg.l_blue.(i);
      incr j
    end
  done;
  if !j <> limg.l_colors then 
    prerr_endline (Printf.sprintf "color match failed %d/%d" !j limg.l_colors);
  let palette' = Array.create !j (-1,-1,-1) in
  Array.blit palette 0 palette' 0 !j;
  for y = 0 to limg.l_height - 1 do
    for x = 0 to limg.l_width - 1 do
      let c = 
	try
	  Char.code limg.l_pixels.[x * limg.l_height + y] 
	with
	  Invalid_argument _ -> prerr_endline (Printf.sprintf "Error happend x=%d y=%d (w=%d h=%d)" x y limg.l_width limg.l_height); 0
      in
      pixels.[y * limg.l_width + x] <- trans.(c); 
    done
  done;
  { width= limg.l_width;
    height= limg.l_height;
    pixels= pixels;
    palette= palette' 
  } 

(* *************************************************** *)
(* LOAD A GIF FILE AND QUANTIZE IT IN ORDER TO MAKE IT *)
(* FITTING THE GIVEN PALETTE pal.                      *)
(* *************************************************** *)
let load_quantize pal name = (* with color quantization *)
  (* use lower interf. *)
  let limg = loadlow name in
  (* translate *)
  let pixels = String.create (limg.l_width * limg.l_height) in
  let nearests = Array.create 256 (-1) in
  for i = 0 to 256 - 1 do
    if not limg.l_free.(i) then begin
      let rgb = limg.l_red.(i),  limg.l_green.(i), limg.l_blue.(i) in
      nearests.(i) <- get_nearest pal rgb;
      if nearests.(i) = -1 then raise (Failure "quantize table init error")
    end
  done;
  for y = 0 to limg.l_height - 1 do
    for x = 0 to limg.l_width - 1 do
      let c = Char.code limg.l_pixels.[x * limg.l_height + y] in
      let qc = try nearests.(c) with
	_ -> prerr_endline (Printf.sprintf
			    "Strange color!!! %d" c); (-1)
      in
      let qc = 
	if qc = (-1) then begin
	  prerr_endline (Printf.sprintf "No trans.tabl (%d)" c);
	  0
	end else qc 
     in
      pixels.[x + limg.l_width * y] <- Char.chr qc
    done
  done;
  { width= limg.l_width;
    height= limg.l_height;
    pixels= pixels;
    palette= Array.copy pal 
  } 

