open Svgalib ;;
open Gd ;;


exception Error of string ;;


type point = { mutable x : float ; mutable y : float } ;;


(* Overwrite with the mouse pointer *)
let draw_mouse_pointer x y color =
 vga_set_color color ;
 vga_draw_line (x-1) y (x+1) y ;
 vga_draw_line x (y-1) x (y+1)
;;


(* Compute vect(pq) . vect (rs) *)
let prod_scal p q r s =
  (q.x -. p.x) *. (s.x -. r.x) +. (q.y -. p.y) *. (s.y -. r.y)
;;


(* Very first image - This one doesn't change - Used to reset all *)
let original_image = String.create 64000
(* Buffer containing the image once warped before going in video memory *)
and warped_buffer = String.create 64000
(* Image used as reference for one transformation stage *)
and current_image = String.create 64000 ;;


let add v1 v2 = { x = v1.x +. v2.x; y = v1.y +. v2.y};;
let sub v1 v2 = { x = v1.x -. v2.x; y = v1.y -. v2.y};;
let mul s v = {x = v.x *. s; y = v.y *. s};;

let vect p1 p2 = sub p2 p1 ;;

let default_color = '\000' ;;

let get_pixel image x y = String.get image (y * 320 + x) ;;
let set_pixel image x y color = String.set image (y * 320 + x) color ;;


let morph pix_from pix_to o s =
 if o = s then
   String.blit current_image 0 warped_buffer 0 64000
 else
  begin
  let w = 320 and h = 200 in
  let os_ = vect o s in
  let os_ = mul 0.2 os_ in
  let s = add o os_ in
  let f = add s (vect o s) in
  let e = add o (vect f o) in
  let y = {x = s.x -. os_.y; y = s.y +. os_.x} in
  let sy = vect s y in
  let sy2 = prod_scal s y s y in
  let sesf = prod_scal s e s f in
  let sosf_sy2 = -. prod_scal s o s f /. sy2 in
  let sf2 = prod_scal s f s f in

  (* bounding box *)
  let c = add y (vect s f) in
  let b = add y (vect s e) in
  let z = sub s sy in
  let d = add z (vect s f) in
  let a = add z (vect s e) in
  let xmax = truncate (max (max a.x b.x) (max c.x d.x)) in
  let ymax = truncate (max (max a.y b.y) (max c.y d.y)) in
  let xmin = truncate (min (min a.x b.x) (min c.x d.x)) in
  let ymin = truncate (min (min a.y b.y) (min c.y d.y)) in
  (* For each point of the destination image in the bounding box... *)
  for xi = max xmin 0 to min xmax (w-1)
   do
    for yi = max ymin 0 to min ymax (h-1)
     do
      let m1 = { x = float xi ; y = float yi } in
      let sm1sy = prod_scal s m1 s y in
      let sm1sf = prod_scal s m1 s f in
      (* Test if we are in the warped square *)
      if -. sy2 <= sm1sy && sm1sy <= sy2 && sesf <= sm1sf && sm1sf <= sf2
      then
       (* Yes we are... *)
       (* Test if we are in the compressed or stretched part, i.e on which *)
       (* side of the limit-line. *)
       if -. sm1sf /. abs_float (sm1sy) <= sosf_sy2 then
        begin
        (* Compress *)
        let fr = { x = sm1sy /. sy2 *. sy.x; y = sm1sy /. sy2 *. sy.y } in
        let r = { x = f.x +. fr.x; y = f.y +. fr.y } in
        let rm1 = { x = m1.x -. r.x; y = m1.y -. r.y } in
        let pp = abs_float (sm1sy /. sy2) in
        let scale = pp +. (1.0 -. pp) *. 0.5 in
        let rm = { x = rm1.x /. scale; y = rm1.y /. scale } in
        let m = { x = rm.x +. r.x; y = rm.y +. r.y } in
        let mxi = truncate m.x in
        let myi = truncate m.y in
        let color = 
	  if mxi < 0 || mxi >= w || myi < 0 || myi >= h then default_color
	  else get_pixel pix_from mxi myi in
        set_pixel pix_to xi yi color
        end
       else
        begin
        (* Stretch *)
        let ep = { x = sm1sy /. sy2 *. sy.x; y = sm1sy /. sy2 *. sy.y } in
        let p = { x = e.x +. ep.x; y = e.y +. ep.y } in
        let pm1 = { x = m1.x -. p.x; y = m1.y -. p.y } in
        let pp = abs_float (sm1sy /. sy2) in
        let scale = pp +. (1.0 -. pp) *. 1.5 in
        let pm = { x = pm1.x /. scale; y = pm1.y /. scale } in
        let m = { x = pm.x +. p.x; y = pm.y +. p.y } in
        let mxi = truncate m.x in
        let myi = truncate m.y in
        let color = 
	  if mxi < 0 || mxi >= w || myi < 0 || myi >= h then default_color
	  else get_pixel pix_from mxi myi in
        set_pixel pix_to xi yi color
        end
      else
       begin
       (* We aren't in the square so transformation doesn't affect the point *)
       let color = String.get current_image (yi * 320 + xi) in
       String.set warped_buffer (yi * 320 + xi) color
       end
     done
   done
 end
;;


(* Load a 320x200x8bits GIF color image *)
let load_gif file_name buffer =
  let im = load file_name in
  if im.width <> 320 then raise (Error "image width <> 320") ;
  if im.height <> 200 then raise (Error "image height <> 200") ;
  String.blit im.pixels 0 buffer 0 64000 ;
  for i = 0 to (Array.length im.palette) - 1
   do
    let (r,g,b) = im.palette.(i) in
      vga_setpalette_entry i (r/4) (g/4) (b/4)
   done
;;


let main () =
 let gif_name = ref "" in
 Arg.parse []
           (fun n -> gif_name := n)
           "Caml-warp v0.1\nUsage :\n\tcaml-morph <image name (GIF)>\n" ;
 let o = { x = 0.0 ; y = 0.0 } in
 vga_set_mouse_support MOUSE_ENABLE ;
 vga_lores_init () ;
 (try
  load_gif !gif_name original_image
  with GifError msg -> raise (Error msg)) ;
 (* Initialize transformation reference image *)
 String.blit original_image 0 current_image 0 64000 ;
 (* Initialize transformed image *)
 String.blit original_image 0 warped_buffer 0 64000 ;
 (* First screen shot *)
 vga_blit_buffer warped_buffer ;
 vga_mouse_set_position 159 99 ;
 draw_mouse_pointer 159 99 0 ;
 let set_center = ref true in
 while true
  do
  if vga_mouse_update () then
   begin
   let x_s = vga_mouse_getx () and y_s = vga_mouse_gety () in
   begin
   match vga_mouse_get_button () with
    | RIGHT ->
        (* Stop all *)
        vga_lores_close () ;
        exit 0
    | NO_BUTTONPRESS ->
        (* No button pressed or all buttons released *)
        (* Transformation must stop until a new deformation center is *)
        (* defined with the left button. *)
        set_center := true ;            (* We must set a new center *)
        vga_blit_buffer warped_buffer ; (* No changes in image *)
        (* Now, the referenxe image for the next deformation stage is *)
        (* the image in the state it has been left. *)
        String.blit warped_buffer 0 current_image 0 64000 ;
    | LEFT ->
       (* If it's the first click since it was released, that means *)
       (* we have to set the center to the current mouse position *)
       if !set_center then (o.x <- (float x_s) ;
                             o.y <- (float y_s) ;
                             set_center := false) ;
       (* Because left mouse is down, any mouse move will produce a *)
       (* warping on the image. *)
       morph current_image warped_buffer o { x = float x_s ; y = float y_s } ;
       (* Once the image has been warped in the buffer, transfer *)
       (* it to the video RAM. *)
       vga_blit_buffer warped_buffer
    | MIDDLE ->
       (* Reset to initially loaded image *)
       String.blit original_image 0 current_image 0 64000 ;
       String.blit original_image 0 warped_buffer 0 64000 ;
       set_center := true
    | _ ->
       (* Ignore cases where several buttons are pressed simultaneously *)
       vga_blit_buffer warped_buffer
   end ;
   draw_mouse_pointer x_s y_s 0 ;
   end
  done
;;


try main () with
 | Error msg ->
    vga_lores_close () ;
    prerr_string ("Caml-warp error : "^msg^"\n") ;
    exit 0
;;
